@@ -38,6 +38,9 @@ static void gs_window_class_init (GSWindowClass *klass);
 static void gs_window_init       (GSWindow      *window);
 static void gs_window_finalize   (GObject       *object);
 
+static gboolean popup_dialog_idle (GSWindow *window);
+static gboolean gs_window_request_unlock_idle (GSWindow *window);
+
 enum {
         DIALOG_RESPONSE_CANCEL,
         DIALOG_RESPONSE_OK
@@ -220,6 +223,17 @@ gs_window_move_resize_window (GSWindow *window,
                                    window->priv->geometry.height);
 }
 
+static void
+gs_window_real_unrealize (GtkWidget *widget)
+{
+        g_signal_handlers_disconnect_by_func (gtk_window_get_screen (GTK_WINDOW (widget)),
+                                              screen_size_changed,
+                                              widget);
+
+        if (GTK_WIDGET_CLASS (parent_class)->unrealize)
+                GTK_WIDGET_CLASS (parent_class)->unrealize (widget);
+}
+
 static void
 gs_window_real_realize (GtkWidget *widget)
 {
@@ -247,6 +261,68 @@ watchdog_timer (GSWindow *window)
         return TRUE;
 }
 
+static void
+remove_watchdog_timer (GSWindow *window)
+{
+        if (window->priv->watchdog_timer_id != 0) {
+                g_source_remove (window->priv->watchdog_timer_id);
+                window->priv->watchdog_timer_id = 0;
+        }
+}
+
+static void
+add_watchdog_timer (GSWindow *window,
+                    glong     timeout)
+{
+        window->priv->watchdog_timer_id = g_timeout_add (timeout,
+                                                         (GSourceFunc)watchdog_timer,
+                                                         window);
+}
+
+static void
+remove_popup_dialog_idle (GSWindow *window)
+{
+        if (window->priv->popup_dialog_idle_id != 0) {
+                g_source_remove (window->priv->popup_dialog_idle_id);
+                window->priv->popup_dialog_idle_id = 0;
+        }
+}
+
+static void
+add_popup_dialog_idle (GSWindow *window)
+{
+        window->priv->popup_dialog_idle_id = g_idle_add ((GSourceFunc)popup_dialog_idle, window);
+}
+
+static void
+remove_request_unlock_idle (GSWindow *window)
+{
+        if (window->priv->request_unlock_idle_id != 0) {
+                g_source_remove (window->priv->request_unlock_idle_id);
+                window->priv->request_unlock_idle_id = 0;
+        }
+}
+
+static void
+add_request_unlock_idle (GSWindow *window)
+{
+        window->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, window);
+}
+
+static gboolean
+emit_deactivated_idle (GSWindow *window)
+{
+        g_signal_emit (window, signals [DEACTIVATED], 0);
+
+        return FALSE;
+}
+
+static void
+add_emit_deactivated_idle (GSWindow *window)
+{
+        g_idle_add ((GSourceFunc)emit_deactivated_idle, window);
+}
+
 static void
 gs_window_real_show (GtkWidget *widget)
 {
@@ -262,11 +338,8 @@ gs_window_real_show (GtkWidget *widget)
                 g_timer_destroy (window->priv->timer);
         window->priv->timer = g_timer_new ();
 
-        if (window->priv->watchdog_timer_id != 0)
-                g_source_remove (window->priv->watchdog_timer_id);
-        window->priv->watchdog_timer_id = g_timeout_add (30000,
-                                                         (GSourceFunc)watchdog_timer,
-                                                         window);
+        remove_watchdog_timer (window);
+        add_watchdog_timer (window, 30000);
 }
 
 void
@@ -284,8 +357,7 @@ gs_window_real_hide (GtkWidget *widget)
 
         window = GS_WINDOW (widget);
 
-        if (window->priv->watchdog_timer_id != 0)
-                g_source_remove (window->priv->watchdog_timer_id);
+        remove_watchdog_timer (window);
 
         if (GTK_WIDGET_CLASS (parent_class)->hide)
                 GTK_WIDGET_CLASS (parent_class)->hide (widget);
@@ -307,14 +379,6 @@ gs_window_get_gdk_window (GSWindow *window)
         return GTK_WIDGET (window)->window;
 }
 
-static gboolean
-emit_deactivated_idle (GSWindow *window)
-{
-        g_signal_emit (window, signals [DEACTIVATED], 0);
-
-        return FALSE;
-}
-
 static gboolean
 spawn_on_window (GSWindow *window,
                  char     *command,
@@ -481,6 +545,11 @@ static void
 socket_destroyed (GtkWidget *widget,
                   GSWindow  *window)
 {
+        g_signal_handlers_disconnect_by_func (window->priv->socket, socket_show, window);
+        g_signal_handlers_disconnect_by_func (window->priv->socket, socket_destroyed, window);
+        g_signal_handlers_disconnect_by_func (window->priv->socket, plug_added, window);
+        g_signal_handlers_disconnect_by_func (window->priv->socket, plug_removed, window);
+
         window->priv->socket = NULL;
 }
 
@@ -536,10 +605,10 @@ gs_window_dialog_finish (GSWindow *window)
                 int exit_status;
                         
                 exit_status = wait_on_child (window->priv->pid);
-        }
 
-        g_spawn_close_pid (window->priv->pid);
-        window->priv->pid = 0;
+                g_spawn_close_pid (window->priv->pid);
+                window->priv->pid = 0;
+        }
 
         /* remove events for the case were we failed to show socket */
         remove_key_events (window);
@@ -637,7 +706,7 @@ command_watch (GIOChannel   *source,
                 gs_window_dialog_finish (window);
 
                 if (window->priv->dialog_response == DIALOG_RESPONSE_OK) {
-                        g_idle_add ((GSourceFunc)emit_deactivated_idle, window);
+                        add_emit_deactivated_idle (window);
                 }
 
                 gs_window_clear (window);
@@ -649,6 +718,7 @@ command_watch (GIOChannel   *source,
                 window->priv->last_y = -1;
 
                 window->priv->watch_id = 0;
+
                 return FALSE;
         }
 
@@ -695,6 +765,7 @@ popup_dialog_idle (GSWindow *window)
                 command = g_string_append (command, " --enable-logout");
                 g_string_append_printf (command, " --logout-command='%s'", window->priv->logout_command);
         }
+
         if (is_user_switch_enabled (window)) {
                 command = g_string_append (command, " --enable-switch");
         }
@@ -708,7 +779,7 @@ popup_dialog_idle (GSWindow *window)
                                   (GIOFunc)command_watch,
                                   window,
                                   &window->priv->watch_id);
-        if (!result)
+        if (! result)
                 g_warning ("Could not start command: %s", command->str);
 
         g_string_free (command, TRUE);
@@ -727,12 +798,13 @@ gs_window_request_unlock (GSWindow *window)
                 return;
 
         if (! window->priv->lock_enabled) {
-                g_idle_add ((GSourceFunc)emit_deactivated_idle, window);
+                add_emit_deactivated_idle (window);
+
                 return;
         }
 
         if (window->priv->popup_dialog_idle_id == 0) {
-                window->priv->popup_dialog_idle_id = g_idle_add ((GSourceFunc)popup_dialog_idle, window);
+                add_popup_dialog_idle (window);
         }
 
         g_signal_emit (window, signals [DIALOG_UP], 0);
@@ -939,7 +1011,7 @@ gs_window_real_key_press_event (GtkWidget   *widget,
         /* if we don't already have a socket then request an unlock */
         if (! GS_WINDOW (widget)->priv->socket) {
                 if (GS_WINDOW (widget)->priv->request_unlock_idle_id == 0) {
-                        GS_WINDOW (widget)->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
+                        add_request_unlock_idle (GS_WINDOW (widget));
                 }
 
                 catch_events = TRUE;
@@ -986,7 +1058,7 @@ gs_window_real_motion_notify_event (GtkWidget      *widget,
                 /* if we don't already have a socket then request an unlock */
                 if (! window->priv->socket
                     && (window->priv->request_unlock_idle_id == 0)) {
-                        window->priv->request_unlock_idle_id = g_idle_add ((GSourceFunc)gs_window_request_unlock_idle, widget);
+                        add_request_unlock_idle (window);
                 }
                 window->priv->last_x = -1;
                 window->priv->last_y = -1;
@@ -1047,6 +1119,7 @@ gs_window_class_init (GSWindowClass *klass)
         widget_class->show                = gs_window_real_show;
         widget_class->hide                = gs_window_real_hide;
         widget_class->realize             = gs_window_real_realize;
+        widget_class->unrealize           = gs_window_real_unrealize;
         widget_class->key_press_event     = gs_window_real_key_press_event;
         widget_class->motion_notify_event = gs_window_real_motion_notify_event;
         widget_class->size_request        = gs_window_real_size_request;
@@ -1161,6 +1234,15 @@ gs_window_init (GSWindow *window)
                                | GDK_LEAVE_NOTIFY_MASK);
 }
 
+static void
+remove_command_watch (GSWindow *window)
+{
+        if (window->priv->watch_id != 0) {
+                g_source_remove (window->priv->watch_id);
+                window->priv->watch_id = 0;
+        }
+}
+
 static void
 gs_window_finalize (GObject *object)
 {
@@ -1175,27 +1257,24 @@ gs_window_finalize (GObject *object)
 
         g_free (window->priv->logout_command);
 
-        if (window->priv->watchdog_timer_id != 0) {
-                g_source_remove (window->priv->watchdog_timer_id);
-                window->priv->watchdog_timer_id = 0;
-        }
-
-        if (window->priv->request_unlock_idle_id != 0) {
-                g_source_remove (window->priv->request_unlock_idle_id);
-                window->priv->request_unlock_idle_id = 0;
-        }
-
-        if (window->priv->popup_dialog_idle_id != 0) {
-                g_source_remove (window->priv->popup_dialog_idle_id);
-                window->priv->popup_dialog_idle_id = 0;
-        }
+        remove_watchdog_timer (window);
+        remove_request_unlock_idle (window);
+        remove_popup_dialog_idle (window);
 
         if (window->priv->timer)
                 g_timer_destroy (window->priv->timer);
 
-        /* just in case they weren't removed */
         remove_key_events (window);
 
+        remove_command_watch (window);
+
+        /* If a dialog is up we need to signal it
+           and wait on it */
+        if (window->priv->pid > 0) {
+                signal_pid (window->priv->pid, SIGTERM);
+        }
+        gs_window_dialog_finish (window);
+
         G_OBJECT_CLASS (parent_class)->finalize (object);
 }
 
@@ -1207,6 +1286,7 @@ gs_window_new (GdkScreen *screen,
         GObject     *result;
 
         result = g_object_new (GS_TYPE_WINDOW,
+                               "type", GTK_WINDOW_POPUP,
                                "screen", screen,
                                "monitor", monitor,
                                "lock-enabled", lock_enabled,
